#include <i2cmaster.h>
#include <twimaster.c>

#define EEPROM_ADDRESS (byte)0xA0

// There are now three-levels of I2C EEPROM access functions:
// 8-bit byte, 16-bit int, and multi-byte block

void i2c_eeprom_read_block(unsigned int eeaddress, byte *buffer, byte length)
{
  i2c_start_wait(EEPROM_ADDRESS + I2C_WRITE);
  i2c_write( ((byte*)(&eeaddress))[1] ); // msb
  i2c_write( ((byte*)(&eeaddress))[0] ); // lsb
  i2c_rep_start(EEPROM_ADDRESS + I2C_READ);
  for (byte i = 0; i < length - 1; i++)
    buffer[i] = i2c_readAck();
  buffer[length] = i2c_readNak();
  i2c_stop();
}
void i2c_eeprom_write_block(unsigned int eeaddress, byte* buffer, byte length)
{
  i2c_start_wait(EEPROM_ADDRESS + I2C_WRITE);
  i2c_write( ((byte*)(&eeaddress))[1] ); // msb
  i2c_write( ((byte*)(&eeaddress))[0] ); // lsb
  for (byte i = 0; i < length; i++)
    i2c_write(buffer[i]);
  i2c_stop();
  delayMicroseconds(3000);
}

// I2C EEPROM functions for reading and writing 16-bit integers
unsigned int i2c_eeprom_read_int(unsigned int eeaddress)
{
  unsigned int i;
  byte* low = (byte*) &i;
  byte* high = ((byte*) &i) + 1;
  *low = i2c_eeprom_read_byte(eeaddress);
  *high = i2c_eeprom_read_byte(eeaddress + 1);
  return i;
}
void i2c_eeprom_write_int(unsigned int eeaddress, unsigned int data)
{
  i2c_eeprom_write_byte(eeaddress,     ((byte*)(&data))[0] ); // lsb
  i2c_eeprom_write_byte(eeaddress + 1, ((byte*)(&data))[1] ); // msb
}

// I2C EEPROM functions for reading and writing 8-bit bytes
byte i2c_eeprom_read_byte(unsigned int eeaddress)
{
  byte rdata = 0xFF;
  i2c_start_wait(EEPROM_ADDRESS + I2C_WRITE);
  i2c_write( ((byte*)(&eeaddress))[1] ); // msb
  i2c_write( ((byte*)(&eeaddress))[0] ); // lsb
  i2c_rep_start(EEPROM_ADDRESS + I2C_READ);
  rdata = i2c_readNak();
  i2c_stop();
  return rdata;
}
void i2c_eeprom_write_byte(unsigned int eeaddress, byte data)
{
    i2c_start_wait(EEPROM_ADDRESS + I2C_WRITE);
    i2c_write( ((byte*)(&eeaddress))[1] ); // msb
    i2c_write( ((byte*)(&eeaddress))[0] ); // lsb
    i2c_write(data);
    i2c_stop();
    delayMicroseconds(3000);
}

void i2c_eeprom_setup()
{
  i2c_init();
}
