// This is the most complicated section of code.
// It handles communication over USB for uploading new words using the Python program.

#define INPUT_BUFFER_SIZE 32

// TODO - one last todo to remind us to check the usb vID and pID for the library!

// This is the main USB function. Returns when the plug is pulled.
void usb_monitor()
{
  char buf[INPUT_BUFFER_SIZE];
  
  lcd.clear();
  lcd.print("usb mode");
  
  UsbStreamMultibyte.begin();
  while (digitalRead(USB_VUSB) == HIGH)
  {
    UsbStreamMultibyte.refresh();
    if (UsbStreamMultibyte.available())
    {
      UsbStreamMultibyte.read(buf, INPUT_BUFFER_SIZE);
      buf[INPUT_BUFFER_SIZE - 1] = '\0'; // TODO do we need this?
 
      // this is a good debug tool:
      //lcd.clear();
      //for (byte c = 0; c < 16; c++)
        //lcd.print(buf[c]);

      parse_command(buf);

      UsbStreamMultibyte.write(buf, INPUT_BUFFER_SIZE);
    }
  }
  // if they remove the plug, we are done
  UsbStreamMultibyte.end();
  
  lcd.clear();
  lcd.print("done with usb");
}

// usb serial api:
#define ACTION_CLEAR                    'x'
#define ACTION_ADD_CAT                  'c'
#define ACTION_ADD_WORD                 'a'
#define ACTION_GET_NUM_CAT              'n'
#define ACTION_GET_NUM_WORDS            'w'
#define ACTION_GET_CAT_TITLE            't'
#define ACTION_GET_WORD                 'f'

void parse_command(char* buf)
{
  unsigned int cID;
  unsigned int wID;
  
  switch (buf[0])
  {
    case ACTION_CLEAR:
      // x                      Clear all wordlists
      lcd.clear();
      output_string(CLEARING_STR, 0, 1);
      set_category_count(0);
      break;

    case ACTION_ADD_CAT:
      // cCATEGORYTITLE       Add new category, CATEGORYTITLE, max of 15 chars here // TODO how many chars can we handle?
      lcd.clear();
      output_string(ADDING_CATEGORY_STR, 0, 1);
      lcd.setCursor(0, 1);
      lcd.print(buf + 1);
      append_category(buf + 1, 250); // TODO remove this magic number (and max word count) eventually
      break;
      
    case ACTION_ADD_WORD:
      // aCNEWWORDXXXXXXXX   Add new word in category C
      cID = buf[1];
      append_word(cID, buf + 2);
      lcd.clear();
      lcd.print("adding word ");
      lcd.print(cID);
      lcd.setCursor(0, 1);
      lcd.print(buf + 2);
      break;

    case ACTION_GET_NUM_CAT:
      // n                      Get number of categories
      buf[0] = get_category_count();
      lcd.clear();
      lcd.print("get num cat");
      lcd.setCursor(0, 1);
      lcd.print(buf[0]);
      break;
      
    case ACTION_GET_NUM_WORDS: // TODO is this working?
      // wC                     Get number of words in category C
      cID = buf[1];
      buf[0] = get_word_count(cID);
      lcd.clear();
      lcd.print("get num words ");
      lcd.print(cID);
      lcd.setCursor(0, 1);
      lcd.print(buf[0], DEC);
      break;

    case ACTION_GET_CAT_TITLE:
      // tC                   Retrieve the title of category C as up to 16 characters
      cID = buf[1];
      get_category_title(cID, buf);
      lcd.clear();
      lcd.print("get cat title ");
      lcd.print(cID);
      lcd.setCursor(0, 1);
      lcd.print(buf[0], DEC);
      break;

    case ACTION_GET_WORD:
      // fCW             Fetch word W from category C as up to 16 characters
      lcd.clear();
      cID = buf[1];
      wID = (buf[2] << 8) + buf[3];
      get_word(cID, wID, buf);
      lcd.clear();
      lcd.print("get word ");
      lcd.print(cID);
      lcd.print(" ");
      lcd.print(wID);
      lcd.setCursor(0, 1);
      lcd.print(buf);
      break;
  }
}

