/* Nerdle, Flashcard/"Stream of words" Firmware
 7/21/2010
 
 If your device doesn't seem to respond to the buttons, use the 
 Nerdle Calibration sketch to find a good threshold for your
 buttons, and change the values directly following this comment.
 
 Uploading instructions:  
 Bootloader mode is used for updating the sketch on the device,
 not updating words.
 
 To upload sketches to your Nerdle, connect the  Nerdle to the PC
 using the mini-b USB cable.  Then, press the reset button on the
 Nerdle.  The screen should go blank, and the device is now in 
 "bootloader mode." 
 
 Upload a sketch using the keyboard shortcut Ctrl-U or using the
 upload button in the top bar.  The sketch will run on the 
 device after it is uploaded.  To upload again, keep the device 
 plugged into USB and press reset again.
 
 If you enter bootloader mode and don't want to upload anything,
 unplug it from USB, and press reset.  The stored program will
 run.
 */

#define UP_THRESHOLD 150
#define DOWN_THRESHOLD 150
#define LEFT_THRESHOLD 150
#define RIGHT_THRESHOLD 150


#include <UsbStreamMultibyte.h> // TODO should we make a better name for this library?
#include <LiquidCrystal.h>
#include <CapSense.h>
#include <avr/sleep.h>
#include <avr/wdt.h>

// TODO what is this variable used for?
volatile boolean f_wdt = 1;

// TODO what is this variable used for?
int nint;


/*
 Nerdle
 http://wayneandlayne.com/nerdle/
 
 Adam Wolf & Matthew Beckler
 Wayne and Layne
 */



/* Hardware Definitions */
// LCD
#define LCD_POWER 0
#define LCD_RS 6
#define LCD_E 5
#define LCD_DB4 10
#define LCD_DB5 11
#define LCD_DB6 12
#define LCD_DB7 13

// USB
#define USB_VUSB 3

// Piezo speaker
#define SPEAKER_PIN 9

// Capacitive-touch buttons
#define BUTTON_LF_PIN 16
#define BUTTON_RT_PIN 7
#define BUTTON_UP_PIN 8
#define BUTTON_DN_PIN 15
#define BUTTON_COUNT 5
#define SEND_PIN 17



/* Button setup */
#define BUTTON_UP 0
#define BUTTON_RT 1
#define BUTTON_LF 2
#define BUTTON_DN 3

//enum?

//button map
//button 1 = start/stop or up
//button 2 = team 2 or right
//button 3 = team 1 or left
//button 4 = category or down
//button 5 = next or enter

/* Gameplay Definitions */
#define INITIAL_BUZZ_DELAY 750 // The initial time between buzzes when a round begins
#define ROUND_LENGTH  60000 // The length of a round in milliseconds

#define LARGEST_STRING_LENGTH 60

#define LCD_COLS 16

/* String Defines  */
#define HELP_N_STR 0
#define HELP_1_STR 1
#define HELP_2_STR 2
#define HELP_3_STR 3
#define HELP_4_STR 4
#define HELP_5_STR 5
#define HELP_6_STR 6
#define HELP_7_STR 7
#define HELP_8_STR 8
#define HELP_9_STR 9
#define ENTERING_USB_MODE_STR 10
#define NO_WORDS_IN_THAT_CATEGORY_STR 11
#define ROUND_OVER_SCORE_STR 12
#define TESTS_COMPLETE_STR 13
#define NO_CATEGORIES_DEFINED_STR 14
#define TEN_SECOND_BUTTON_TEST_BEGINNING_STR 15
#define BUTTON_1_RELEASED_STR 16
#define BUTTON_1_DEPRESSED_STR 17
#define NERDLE_STR 18
#define CHOOSE_CATEGORY_STR 19
#define MAIN_SCREEN_TURN_ON_STR 20
#define BUZZ_STR 21
#define USB_MODE_STR 22
#define WAYNEANDLAYNE_COM_NERDLE_STR 23
#define PAUSED_STR 24
#define EXITING_STR 25
#define CATEGORY_FULL_STR 26
#define NO_MORE_AVAILABLE_CATEGORIES_STR 27
#define ADDING_CATEGORY_STR 28
#define ADDING_WORD_STR 29
#define CLEARING_STR 30
#define SCORE_STR 31


LiquidCrystal lcd(LCD_RS, LCD_E, LCD_DB4, LCD_DB5, LCD_DB6, LCD_DB7); 

//1M between pins w foil pad attached to second pin
CapSense   cs_up = CapSense(SEND_PIN,BUTTON_UP_PIN);
CapSense   cs_right = CapSense(SEND_PIN,BUTTON_RT_PIN);
CapSense   cs_left = CapSense(SEND_PIN,BUTTON_LF_PIN);
CapSense   cs_down = CapSense(SEND_PIN,BUTTON_DN_PIN);

// TODO what does this variable do?
long sleep_time;

//state table:
// i = initializing
// n = new game
// c = choosing category
// r = round
// p = paused
// x = no categories
// s = scoring
// u = usb
char state = 'i';

//void test()
//{
//  lcd_test();
//  button_test();
//  speaker_test();
//  word_test();
//  "Tests complete."
//  output_string(TESTS_COMPLETE_STR, 1, 1);
//}

void lcd_setup()
{
  pinMode(LCD_POWER, OUTPUT);
  digitalWrite(LCD_POWER, HIGH);
  delay(50);

  lcd.begin(16, 2);
  delay(50);
  lcd.clear();
}

void setup()
{
  delay(5);
  lcd_setup();

  i2c_eeprom_setup();
  pinMode(SPEAKER_PIN, OUTPUT);
  //beep(); // TODO turn these back on!
  delay(1000);
  //beep();

  //piezo.begin(3);

  // TODO what does this do?
  // TODO put this in it's own function?
  // CPU Sleep Modes 
  // SM2 SM1 SM0 Sleep Mode
  // 0    0  0 Idle
  // 0    0  1 ADC Noise Reduction
  // 0    1  0 Power-down
  // 0    1  1 Power-save
  // 1    0  0 Reserved
  // 1    0  1 Reserved
  // 1    1  0 Standby(1)
  SMCR &= ~_BV(SE);
  //cbi( SMCR,SE );      // sleep enable, power down mode
  
  SMCR &= ~_BV(SM0);
  //cbi( SMCR,SM0 );     // power down mode
  
  SMCR |= _BV(SM1);
  //sbi( SMCR,SM1 );     // power down mode
  
  SMCR &= ~_BV(SM2);
  //cbi( SMCR,SM2 );     // power down mode

  setup_watchdog(7); //from nightingale sketch

  state = 'n';
  
  randomSeed(analogRead(3));
  // TODO does this actually work? Could/should we put the random seed into eeprom and
  // increment it each time we run/reseed?

  sleep_time = millis() + 30000;
}

// TODO what does this function do? when should we call it?
void system_sleep()
{
  //we should set all ports to input to save power
  
  ADCSRA &= ~_BV(ADEN);
  //cbi(ADCSRA,ADEN);                    // switch Analog to Digitalconverter OFF
  
  set_sleep_mode(SLEEP_MODE_PWR_DOWN); // sleep mode is set here
  sleep_enable();
  sleep_mode();                        // System sleeps here
  sleep_disable();                     // System continues execution here when watchdog timed out
  
  ADCSRA |= _BV(ADEN);
  //sbi(ADCSRA,ADEN);                    // switch Analog to Digitalconverter ON
}

//****************************************************************
// TODO what does all this mean and do?
// 0=16ms, 1=32ms,2=64ms,3=128ms,4=250ms,5=500ms
// 6=1 sec,7=2 sec, 8=4 sec, 9= 8sec
void setup_watchdog(int ii)
{
  byte bb;
  int ww;

  if (ii > 9 )
    ii = 9;

  bb = ii & 7;

  if (ii > 7)
    bb|= (1<<5);

  bb|= (1<<WDCE);

  ww = bb;

  MCUSR &= ~(1<<WDRF);

  // start timed sequence
  WDTCSR |= (1<<WDCE) | (1<<WDE);

  // set new watchdog timeout value
  WDTCSR = bb;
  WDTCSR |= _BV(WDIE);
}

//****************************************************************  

// Watchdog Interrupt Service / is executed when  watchdog timed out

ISR(WDT_vect) {
  f_wdt=1;  // set global flag - TODO what does this flag do?
}


// TODO what do these do?
static char button_was_pushed[BUTTON_COUNT] = {0, 0, 0, 0, 0};
static char button_state[BUTTON_COUNT] =      {0, 0, 0, 0, 0};

// TODO what does this function do, what does it return?
char update_buttons()
{
  char changes;
  changes = 0;

  long total0 =  cs_up.capSense(30);
  long total1 =  cs_right.capSense(30);
  long total2 =  cs_left.capSense(30);
  long total3 =  cs_down.capSense(30);

  //Serial.print("Raw Button Data: ");
  //Serial.print(total0);
  //Serial.print('\t');
  //Serial.print(total1);
  //Serial.print('\t');
  //Serial.print(total2);
  //Serial.print('\t');
  //Serial.println(total3);


  //TODO: make buzzing noise
  if (total0 > UP_THRESHOLD)
  {
    sleep_time = millis() + 30000;
    if (button_state[0] == 1)
    {
      //it is currently pushed, and last time it was pushed.
      button_was_pushed[0] = 0;
    } 
    else
    {
      //it is now pushed and last time it wasn't
      tone(SPEAKER_PIN, 1000, 10);
      button_was_pushed[0] = 1;
      changes = 1;
    }

    button_state[0] = 1;
  }
  else
  {
    if (button_state[0] == 1)
    {
      //we weren't pushed, but now we are!
      tone(SPEAKER_PIN, 500, 10);
      changes = 1;
    }
    button_was_pushed[0] = 0;
    button_state[0] = 0;
  }


  if (total1 > RIGHT_THRESHOLD)
  {
    sleep_time = millis() + 30000;
    if (button_state[1] == 1)
    {
      button_was_pushed[1] = 0;
    } 
    else
    {
      tone(SPEAKER_PIN, 1000, 10);
      changes = 1;
      button_was_pushed[1] = 1;
    }
    button_state[1] = 1;
  }
  else
  {
    if (button_state[1] == 1)
    {
      //we weren't pushed, but now we are!
      tone(SPEAKER_PIN, 500, 10);
      changes = 1;
    }
    button_was_pushed[1] = 0;
    button_state[1] = 0;
  }

  if (total2 > LEFT_THRESHOLD)
  {

    if (button_state[2] == 1)
    {
      button_was_pushed[2] = 0;
    } 
    else
    {
      tone(SPEAKER_PIN, 1000, 10);
      changes = 1;

      button_was_pushed[2] = 1;
    }
    button_state[2] = 1;
  }
  else
  {
    if (button_state[2] == 1)
    {
      //we weren't pushed, but now we are!
      tone(SPEAKER_PIN, 500, 10);
      changes = 1;
    }
    button_was_pushed[2] = 0;
    button_state[2] = 0;
  }

  if (total3 > DOWN_THRESHOLD)
  {
    sleep_time = millis() + 30000;
    if (button_state[3] == 1)
    {
      button_was_pushed[3] = 0;
    } 
    else
    {
      tone(SPEAKER_PIN, 1000, 10);
      changes = 1;
      button_was_pushed[3] = 1;
    }
    button_state[3] = 1;
  }
  else
  {
    if (button_state[3] == 1)
    {
      //we weren't pushed, but now we are!
      tone(SPEAKER_PIN, 500, 10);
      changes = 1;
    }
    button_was_pushed[3] = 0;
    button_state[3] = 0;
  }
  return changes;
}


void loop()
{
  static byte cID = 0;
  static unsigned int wID = 0;
  static int team1_score = -1;
  static int team2_score = -1;
  static byte last_category = 0;
  static long round_start = -1;
  static long round_end = -1;
  static int round_score = -1;
  static int buzz_delay = -1;
  static long last_buzz = -1;
  static int buzz_count = -1;
  static long round_length = -1;
  long current_time = millis();
  static long pause_time = -1;
/*
  // deal with sleeping
  if (millis() > sleep_time) // TODO should this use current_time?
  {
    char changes;
    lcd.clear();
    lcd.print("sleeping!");
    lcd.clear();
    //we should save power here!

    system_sleep();
    //we should do something here so if we do wake up *and get pushed!, reset the state to n, turn on the display.
    while (1)
    {

      changes = update_buttons();
      if (changes)
      {
        lcd_setup();
        state = 'n';
        break;
      }
    }
  }
*/
  update_buttons();


  // detect if usb cable is connected
  if (digitalRead(USB_VUSB) == HIGH)
  {
    // stop whatever we were doing, and go to usb mode!
    state = 'u';
  }

  switch (state)
  {
    case 'n':
      game_startup();
      sleep_time = millis() + 30000;
      //the machine has just been turned on
      delay(50);

      if (get_category_count() == 0)
      {
        lcd.clear();
        lcd.print("No wordlist loaded!"); // TODO add to string list
        // TODO what should we do here?
      } 
      else
      {
        //start picking the categories
        cID = 0;
        state = 'c';
        output_string(CHOOSE_CATEGORY_STR, 1, 1);
        //delay(2000); // TODO why was this here?
        display_category(cID);
        break;
      }
      break;

    case 'c':
      // we're picking categories
      if (button_was_pushed[BUTTON_RT] or button_was_pushed[BUTTON_DN])
      {
        //next category
        cID = next_category(cID);
        display_category(cID);
      }
      else if (button_was_pushed[BUTTON_LF])
      {
        //previous category
        cID = previous_category(cID);
        display_category(cID);
      }
      else if (button_was_pushed[BUTTON_UP])
      {
        //if there are any words in that category, let's start it!
        if (get_word_count(cID) > 0)
        {
          state = 'r';
          round_score = 0;
          round_end = current_time + ROUND_LENGTH;
          init_mixer();
          wID = next_word_index(cID);
  
          last_buzz = current_time;
          buzz_delay = INITIAL_BUZZ_DELAY;
          display_word_with_score(cID, wID, round_score);
          buzz_count = 0;
          break;
        }
        else
        { 
          //no words in that category.
          output_string(NO_WORDS_IN_THAT_CATEGORY_STR, 1, 1);
        }
      }
      else
      {
        // no buttons pressed?
      }
      break;


    case 'r':
      //we're starting a round
  
      sleep_time = millis() + 120000; // basically, don't sleep during an active game, durrr.  
      //If it pauses though, and no one pushes buttons for a long time, it will cut you off.
      // TODO - should we make this line into a macro? kick_dog(); ?
      // TODO how long should the sleep time be?
  
      if (current_time >= round_end)
      {
        //round over!
        beep_end();
        display_end_of_round(round_score);
        delay(2000);
  
        state = 'c';
        output_string(CHOOSE_CATEGORY_STR, 1, 1);
        display_category(cID);
        break;
      }
  
      //should we buzz?
      if (current_time >= last_buzz + buzz_delay)
      { 
        //it's buzz time!
        last_buzz = current_time;
        buzz_count = buzz_count + 1;
        beep_tick();
  
        if (buzz_count > 30)
        {
          buzz_count = 0;
          if (buzz_delay > 100)
          {
            buzz_delay = .75 * buzz_delay;
          }
        }
      }

      if (button_was_pushed[BUTTON_UP])
      {
        //pause!
        output_string(PAUSED_STR, 1, 1);
        pause_time = current_time;
        state = 'p';
        break;
      }
  
      if (button_was_pushed[BUTTON_RT])
      {
        //advance word without incrementing score
        wID = next_word_index(cID);
        display_word_with_score(cID, wID, round_score);
      } 
      else if (button_was_pushed[BUTTON_LF])
      {
        //go back to the last word without incrementing score
        wID = previous_word_index(cID);
        display_word_with_score(cID, wID, round_score);
      } 
      else if (button_was_pushed[BUTTON_DN])
      {
        //advance word and increment score
        round_score++;
        wID = next_word_index(cID);
        display_word_with_score(cID, wID, round_score);
        beep_point();
      }
      break;

    case 'p':
      // currently paused
      if (button_was_pushed[BUTTON_UP])
      {
        // compensate - need to recompute the two times below due to the pause time
        last_buzz = last_buzz + current_time - pause_time;
        round_end = round_end + current_time - pause_time;
        state = 'r';
        display_word_with_score(cID, wID, round_score);
        pause_time = -1;
      }
      break;

    case 'u':
      // go into usb mode
      usb_monitor();
      lcd.clear();
      lcd.print("Restarting...");
      delay(1000);
      state = 'n';
      break;
  
    default:
      break;

  } // end of main state machine switch() statement
  
} // end of loop()

