byte next_category(byte current_category)
{
  byte category_count = get_category_count();
  return (current_category == category_count - 1) ? (0) : (current_category + 1);
}
byte previous_category(byte current_category)
{
  byte category_count = get_category_count();
  return (current_category == 0) ? (category_count - 1) : (current_category - 1);
}

unsigned int next_word_index(byte category)
{
  return word_index_mixer(category, 1);
}
unsigned int previous_word_index(byte category)
{
  return word_index_mixer(category, 0);
}

// This is Matthew's new pseudo random number generator so the order of the words changes each time you play.
//
// Can be used for large and small categories. There would be no disadvantage to using the 16 bit counter for
// a very small category, except that we have to keep trying until we get a value that falls within the
// category size.
// 
// We basically have a counter that we move forwards and backwards as needed. We then apply an XOR and a 
// bit mixing operation to make it look random enough.
byte prnd8, count8;
unsigned int prnd16, count16;
void init_mixer() // TODO should be called after every finished round, I think
{
  prnd8 =  random(1, 0x100); // TODO should this be random(0, 0x100); ?
  count8 = random(1, 0x100);
  prnd16 =  random(1, 0x10000);
  count16 = random(1, 0x10000);
}
unsigned int word_index_mixer(byte category, byte increment)
{
  unsigned int word_count = get_word_count(category);
  unsigned int retval;
  do
  {
    if (word_count <= 256)
    {
      if (increment){count8++;}else{count8--;} // rollover is just fine
      // count8 += increment ? 1 : -1; TODO can we replace the above line with this one?
      retval = count8 ^ prnd8;
      retval = (retval >> 4) | ((retval & 0x0F) << 4);
    } else {
      if (increment){count16++;}else{count16--;} // rollover is just fine
      // count16 += increment ? 1 : -1; TODO can we replace the above line with this one?
      retval = count16 ^ prnd16;
      retval = (retval >> 12) | ((retval & 0x0F00) >> 4) | ((retval & 0x00F0) << 4) | (retval << 12);
    }
  } while (retval >= word_count);
  return retval;
}


//every word has a max length of 15 characters plus one \0 at the end.  we store them at regular places in the program.
//we are assuming the buffer you are passing in is of size MAX_LENGTH 16.

//TODO: I should make category a struct...

#define MEMORY_BLOCK_SIZE 16
#define MAX_WORD_LENGTH 16
#define EEPROM_ADDRESS 0x50
#define CATEGORY_COUNT_ADDR 0
#define MAX_CATEGORY_COUNT 32
//there is a maximum of 255 categories because we use a single byte to store the count.
#define CATEGORY_SIZE 64
#define CATEGORY_SIZE_BYTES 4
#define CATEGORY_WORD_COUNT_OFFSET MEMORY_BLOCK_SIZE
#define CATEGORY_FIRST_WORD_ADDR_OFFSET (2 * MEMORY_BLOCK_SIZE)
#define CATEGORY_MAX_WORD_OFFSET (3 * MEMORY_BLOCK_SIZE)
//everything is done in blocks of 16... there is 1 block for title, 1 block for word starting address, 1 block for word count, 1 block for max word count,
// so 4x16 = 64


//block 0
// category count

//block 1
//category 0 title
//block 2
//category 0 word count
//block 3
//category 0 first word address
//block 4
//category 0 max_word_count


byte get_category_count()
{
  return i2c_eeprom_read_byte(CATEGORY_COUNT_ADDR);
}
void set_category_count(byte count)
{
  i2c_eeprom_write_byte(CATEGORY_COUNT_ADDR, count);
}
void increment_category_count()
{
  set_category_count(get_category_count() + 1);
}

// the address of this category's information chunk, not the location of its words
unsigned int get_category_address(byte cID)
{
  return MEMORY_BLOCK_SIZE + (cID * CATEGORY_SIZE); // the first block is category count
}

// Dealing with category titles
void get_category_title(byte cID, char* buffer)
{
  unsigned int address = get_category_address(cID);
  i2c_eeprom_read_block(address, (byte*) buffer, 16);
}
void set_category_title(byte cID, char* buffer)
{
  unsigned int address = get_category_address(cID);
  i2c_eeprom_write_block(address, (byte*) buffer, 16);
}

// Dealing with category word count
unsigned int get_word_count(byte cID)
{
  int word_count_address = get_category_address(cID) + CATEGORY_WORD_COUNT_OFFSET;
  return i2c_eeprom_read_int(word_count_address);
}
void set_word_count(byte cID, unsigned int count)
{
  int word_count_address = get_category_address(cID) + CATEGORY_WORD_COUNT_OFFSET;
  i2c_eeprom_write_int(word_count_address, count);
}
void increment_word_count(byte cID)
{
  set_word_count(cID, get_word_count(cID) + 1);
}

// Dealing with address of first word
unsigned int get_first_word_address(byte cID)
{
  unsigned int location = get_category_address(cID) + CATEGORY_FIRST_WORD_ADDR_OFFSET;
  return i2c_eeprom_read_int(location);
}
void set_first_word_address(byte cID, unsigned int address)
{
  unsigned int location = get_category_address(cID) + CATEGORY_FIRST_WORD_ADDR_OFFSET;
  i2c_eeprom_write_int(location, address);
}

unsigned int get_word_address(byte cID, unsigned int index)
{
  //the address is the address of the first word in that category, plus MEMORY_BLOCK_SIZE*index
  //get the address of the specific word
  return get_first_word_address(cID) + (index * MEMORY_BLOCK_SIZE);
}

// Dealing with the max word count
unsigned int get_max_word_count(byte cID)
{
  unsigned int max_word_count_address = get_category_address(cID) + CATEGORY_MAX_WORD_OFFSET;
  return i2c_eeprom_read_int(max_word_count_address);
}
void set_max_word_count(byte cID, unsigned int count)
{
  int max_word_count_address = get_category_address(cID) + CATEGORY_MAX_WORD_OFFSET;
  i2c_eeprom_write_int(max_word_count_address, count);
}

// Dealing with individual words
void get_word(byte cID, unsigned int wID, char* buffer)
{
  unsigned int address = get_word_address(cID, wID);
  i2c_eeprom_read_block(address, (byte*) buffer, 16); // TODO magic number, omg
}
void set_word(byte cID, unsigned int wID, char* buffer)
{
  unsigned int address = get_word_address(cID, wID);
  i2c_eeprom_write_block(address, (byte*) buffer, 16); // TODO magic number, omg
}
unsigned int append_word(byte cID, char* buffer)
{
  // TODO can we get rid of the return value here? Who uses the return value? I guess it's used in the api.
  unsigned int max_words = get_max_word_count(cID);
  unsigned int word_count = get_word_count(cID);
  if (max_words == word_count)
  {
    //category full
    lcd.clear();
    lcd.print("category full!");
    return -1;
  } 
  else
  {
    unsigned int address = get_word_address(cID, word_count);
    set_word(cID, word_count, buffer);
    increment_word_count(cID);
    return word_count; // yes, we want the old, pre-incremented value
  }
}

// Dealing with categories
void set_category(byte cID, char* title, unsigned int word_count, unsigned int first_word_address, unsigned int max_word_count)
{
  // this function takes 36 ms!
  set_category_title(cID, title);
  set_word_count(cID, word_count);
  set_first_word_address(cID, first_word_address);
  set_max_word_count(cID, max_word_count);
}
int append_category(char* title, unsigned int max_word_count)
{
  unsigned int category_count, first_word_address;
  unsigned int word_count = 0;
  category_count = get_category_count();
  if (category_count == MAX_CATEGORY_COUNT)
  {
    //too many categories
    lcd.clear();
    lcd.print("too many cats");
    return -1;
  }
  else if (category_count == 0)
  {
    first_word_address = MEMORY_BLOCK_SIZE + (MAX_CATEGORY_COUNT * CATEGORY_SIZE); // remember, the first block is current category count...
  }
  else 
  {
    //we need to find the previous category's last possible word slot, and then go one past that.  
    //but don't forget the first block is current category count...
    first_word_address = MEMORY_BLOCK_SIZE + get_word_address(category_count - 1, get_max_word_count(category_count - 1));
  }
  
  //at this point, we know the first_word_address, and everything else has been passed in!
  //we also know that we're not out of categories, or you would have left here already.
  
  increment_category_count();
  set_category(category_count, title, word_count, first_word_address, max_word_count);
}
  
void display_category(int cID)
{
  char buf[33]; // TODO fix magic number 33?
  strcpy_P(buf, (char*)pgm_read_word(&(string_table[CHOOSE_CATEGORY_STR])));
  get_category_title(cID, buf+16);
  output_string(buf, 1, 1);
}

